@@ -25,6 +25,8 @@
 #include "mirrordir.h"
 #include "vfs/vfs.h"
 #include "cmdlineopt.h"
+#include "netrc.h"
+#include "mad.h"
 
 #ifdef HAVE_GETTIMEOFDAY
 #define NICE_MESS "[--nice <num>]"
@@ -67,11 +69,19 @@ int backup_howmany = 0;
 int copy_mode = 0;
 int allow_chmod = 1;
 int allow_chown = 1;
+int allow_netrc = 0; /*EK*/
 int ignore_size = 0;
+extern int ftpfs_option_allow_empty_directories;
+
 char *ftpfs_password = 0;
 unsigned long allow_permissions_mask = 0xFFFFFFFFL;
 time_t outdate_time = 0;
 
+char **control_mirror_directories = 0;
+
+void *free_me[256];
+unsigned char free_num = 0;
+
 /* internal errno */
 #define DIRECTORY_NOT_CREATED 1
 int mirrordir_errno = 0;
@@ -103,6 +113,7 @@ int block_pos = BLOCK_NOT_USED;
 
 int file_in_use = 0;
 int real_bad_error = 0;
+int ignore_next_exclude = 0;
 
 struct directory_list {
     struct stat stat;
@@ -117,6 +128,7 @@ struct exclude_list {
 #define EXCLUDE_GLOB		2
 #define EXCLUDE_LITERAL		3
     int type;
+    int ignore;		/* file/path not to be excluded, but to be overlooked completely */
     regex_t regexp;
     struct exclude_list *next;
     struct exclude_list *prev;
@@ -235,48 +247,54 @@ struct directory_list *read_list (char *path)
 {
     struct directory_list *dl = 0, *next;
     DIR *dir;
+    struct dirent *r;
 
     dir = mc_opendir (path);
     if (!dir) {
 	progmess_strerror ("unable to open directory", path);
 	return 0;
     }
-    for (;;) {
+    dl = malloc (sizeof (struct directory_list));
+    memset (dl, 0, sizeof (struct directory_list));
+    dl->name = strdup ("..");
+
+    do {
 	int l;
-	struct dirent *r;
 	r = mc_readdir (dir);
 	check_interrupt ();
-	if (!r)			/* done */
-	    break;
 	next = malloc (sizeof (struct directory_list));
 	memset (next, 0, sizeof (struct directory_list));
 	next->prev = dl;
 	if (dl)
 	    dl->next = next;
 	dl = next;
-	l = NAMLEN (r);
-	dl->name = malloc (l + 1);
-	memcpy (dl->name, r->d_name, l);
-	dl->name[l] = '\0';
-    }
+	if (r) {
+	    l = NAMLEN (r);
+	    dl->name = malloc (l + 1);
+	    memcpy (dl->name, r->d_name, l);
+	    dl->name[l] = '\0';
+	}
+    } while (r);
 
 /* go to start of list, list will not be empty because of . and .. */
     rewind (dl);
 
     for (;;) {
-	if (strcmp ("..", dl->name) && strcmp (".", dl->name)) {
-	    char *p;
-	    p = join (path, dl->name);
-	    if (mc_lstat (p, &dl->stat)) {
-		progmess_strerror ("error trying to stat file", p);
+	if (dl->name) {
+	    if (strcmp ("..", dl->name) && strcmp (".", dl->name)) {
+		char *p;
+		p = join (path, dl->name);
+		if (mc_lstat (p, &dl->stat)) {
+		    progmess_strerror ("error trying to stat file", p);
+		    free (dl->name);
+		    dl->name = 0;
+		}
+		check_interrupt ();
+		free (p);
+	    } else {
 		free (dl->name);
 		dl->name = 0;
 	    }
-	    check_interrupt ();
-	    free (p);
-	} else {
-	    free (dl->name);
-	    dl->name = 0;
 	}
 	if (!dl->next)
 	    break;
@@ -578,6 +596,8 @@ void copy_regular_file (char *p, char *q, struct stat *stat)
     char buf[COPY_BUF_SIZE];
     off_t togo;
 
+    check_interrupt ();
+
     set_vfs_temp_dir (p, q);
 
     if (verbose) {
@@ -599,7 +619,10 @@ void copy_regular_file (char *p, char *q, struct stat *stat)
    will return an error and we are safe.
  */
     if (f < 0) {
-	progmess_strerror ("unable to open control file for writing", p);
+	if (only_read)
+	    progmess_strerror ("unable to open control file for reading", p);
+	else
+	    progmess_strerror ("unable to open control file for reading and writing", p);
 	if (!only_read) {
 	    file_in_use++;
 	    real_bad_error++;
@@ -621,7 +644,6 @@ void copy_regular_file (char *p, char *q, struct stat *stat)
 	    off_t count, c;
 	    char *b;
 	    be_nice_to_cpu ();
-	    check_interrupt ();
 	    count = mc_read (f, buf, COPY_BUF_SIZE);
 	    if (count < 0) {
 		progmess_strerror ("error trying to read from file", p);
@@ -633,7 +655,6 @@ void copy_regular_file (char *p, char *q, struct stat *stat)
 	    togo -= count;
 	    while (count > 0) {
 		c = mc_write (g, b, count);
-		check_interrupt ();
 		if (c < 0) {
 		    progmess_strerror ("error trying to write to file", q);
 		    mc_close (f);
@@ -663,6 +684,7 @@ void copy_regular_file (char *p, char *q, struct stat *stat)
     if (rest_access)
 	utime_check (stat->st_mtime, stat->st_atime, p, 1);
     utime_check (stat->st_mtime, stat->st_atime, q, 0);
+    check_interrupt ();
 }
 
 int copy_symlink (char *p, char *q, int force)
@@ -964,18 +986,26 @@ int duplicate (char *p, char *q, struct directory_entry *d, struct directory_lis
     return 0;
 }
 
+enum {
+    PATH_INCLUDED,
+    PATH_IGNORED,
+    PATH_EXCLUDED,
+};
+
 int excluded (char *p)
 {
     char *q;
     struct exclude_list *l;
     if (is_listed (p))
-	return 1;
+	return PATH_EXCLUDED;
     if (!exclude_list)
-	return 0;
+	return PATH_INCLUDED;
     rewind (exclude_list);
     q = strrchr (p, '/') + 1;	/* full paths, so there will always be a / somewhere */
     l = exclude_list;
     while (l) {
+	int ignore;
+	ignore = l->ignore ? PATH_IGNORED : PATH_EXCLUDED;
 	switch (l->type) {
 	case EXCLUDE_LITERAL:
 	    if (!strcmp (p, l->name)) {		/* it only occurs once, so we can remove it */
@@ -983,7 +1013,7 @@ int excluded (char *p)
 		if (!l->prev && !l->next) {
 		    exclude_list = 0;
 		    free (l);
-		    return 1;
+		    return ignore;
 		}
 		if (!l->prev) {
 		    l->next->prev = 0;
@@ -996,21 +1026,21 @@ int excluded (char *p)
 		if ((unsigned long) l == (unsigned long) exclude_list)
 		    exclude_list = l->next;
 		free (l);
-		return 1;
+		return ignore;
 	    }
 	    break;
 	case EXCLUDE_REGEXP:
 	    if (!regexec (&l->regexp, p, 0, NULL, 0))
-		return 1;
+		return ignore;
 	    break;
 	case EXCLUDE_GLOB:
 	    if (!regexec (&l->regexp, q, 0, NULL, 0))
-		return 1;
+		return ignore;
 	    break;
 	}
 	l = l->next;
     }
-    return 0;
+    return PATH_INCLUDED;
 }
 
 unsigned long blocks_to_kilobytes (unsigned long blocks)
@@ -1164,13 +1194,18 @@ int recurs_dir (char *control, char *mirror,...)
 	}
 	if (dl->name) {
 	    char *p, *q;
+	    int e;
 	    check_interrupt ();
 	    if (control) {
 		p = join (control, dl->name);
 	    } else {
 		p = strdup (*files);
 	    }
-	    if (!excluded (p) && (mode_t_to_unsigned_long (dl->stat.st_mode) & allow_permissions_mask)) {	/* excluded files must not count toward the running total of blocks */
+	    e = excluded (p);
+	    if (e == PATH_IGNORED) {
+		/* ignoring this entry, so remove it from the mirror list */
+		remove_name_entry (dm, dl->name);
+	    } else if (e == PATH_INCLUDED && (mode_t_to_unsigned_long (dl->stat.st_mode) & allow_permissions_mask)) {	/* excluded files must not count toward the running total of blocks */
 		if (block_range (p, (struct directory_entry *) dl)) {
 		    struct directory_list *d;
 		    int m;
@@ -1309,10 +1344,12 @@ Copydir is equivalent to\n\
   -M, --max-bytes <num>[K|M|G]  max bytes to mirror, then abort with filename\n\
   -s, --starting-file <path>    start mirroring only after reading <path>\n\
 \n\
-  -X, --exclude <path>          exclude full path matching <path>\n\
-  -F, --exclude-from <file>     read list of files to exclude from <file>\n\
-  -G, --exclude-glob <expr>     exclude file/directory names matching <expr>\n\
-  -R, --exclude-regexp <expr>   exclude full paths matching <expr>\n\
+  [-i, --ignore-next-exclude]        the next exclude option means to leave\n\
+                                     the path completely unchanged\n\
+  [-i] -X, --exclude <path>          exclude full path matching <path>\n\
+  -F, --exclude-from <file>          read list of files to exclude from <file>\n\
+  [-i] -G, --exclude-glob <expr>     exclude file/dir names matching <expr>\n\
+  [-i] -R, --exclude-regexp <expr>   exclude full paths matching <expr>\n\
 \n\
   -T, --allow-matches <type>    only mirror files matching <type> symbolic\n\
                                 expression.\n\
@@ -1328,6 +1365,10 @@ Copydir is equivalent to\n\
   -o, --only-read               don't open control files for writing (default)\n\
 \n\
   -p, --password <password>     ftp password\n\
+      --netrc                   consult ~/.netrc (this is the default)\n\
+      --no-netrc                don't consult ~/.netrc\n\
+\n\
+      --allow-empty-ftp-dirs    some servers cause an error with no . nor ..\n\
 \n\
   -t, --dry-run, --test-only    output merely what would be done\n\
 \n\
@@ -1346,20 +1387,28 @@ Please send comments, suggestions and bug reports to\n\
 
 char cwd[MAX_PATH_LEN + 1];
 
+#ifdef HAVE_MAD
+char *mad_pathdup (char *p, char *file, int line)
+#else
 char *pathdup (char *p)
+#endif
 {
-    char *q, *r;
+    char *q, *r, *f = 0;
 
     if (!is_special_prefix (p)) {
 	if (*p != '/') {
-	    r = malloc (strlen (cwd) + strlen (p) + 2);
+	    f = r = malloc (strlen (cwd) + strlen (p) + 2);
 	    strcpy (r, cwd);
 	    strcat (r, "/");
 	    strcat (r, p);
 	    p = r;
 	}
     }
+#ifdef HAVE_MAD
+    r = q = mad_alloc (strlen (p) + 2, file, line);
+#else
     r = q = malloc (strlen (p) + 2);
+#endif
     if (is_special_prefix (p)) {
 	while (*p != ':')
 	    *q++ = *p++;
@@ -1399,6 +1448,8 @@ char *pathdup (char *p)
     if (r[0] && r[1])
 	if (*--q == '/')
 	    *q = '\0';
+    if (f)
+	free (f);
     return r;
 }
 
@@ -1463,6 +1514,8 @@ int opt_exclude_callback (int a, char *s)
     case 'X':
 	exclude_list = add_to_exclude_list (exclude_list, s, 1);
 	exclude_list->type = EXCLUDE_LITERAL;
+	exclude_list->ignore = (ignore_next_exclude != 0);
+	ignore_next_exclude = 0;
 	break;
     case 'G':
 	exclude_list = add_to_exclude_list (exclude_list, s, 0);
@@ -1470,6 +1523,8 @@ int opt_exclude_callback (int a, char *s)
 	free (exclude_list->name);
 	exclude_list->name = p;
 	exclude_list->type = EXCLUDE_GLOB;
+	exclude_list->ignore = (ignore_next_exclude != 0);
+	ignore_next_exclude = 0;
 	if (regcomp (&exclude_list->regexp, exclude_list->name, REG_EXTENDED | REG_NOSUB)) {
 	    progmess ("error converting glob expression to regular expression", s);
 	    return 1;
@@ -1478,6 +1533,8 @@ int opt_exclude_callback (int a, char *s)
     case 'R':
 	exclude_list = add_to_exclude_list (exclude_list, s, 0);
 	exclude_list->type = EXCLUDE_REGEXP;
+	exclude_list->ignore = (ignore_next_exclude != 0);
+	ignore_next_exclude = 0;
 	if (regcomp (&exclude_list->regexp, exclude_list->name, REG_EXTENDED | REG_NOSUB)) {
 	    progmess ("error compiling regular expression", s);
 	    return 1;
@@ -1542,6 +1599,7 @@ int calc_bytes (char *s)
     return 0;
 }
 
+
 int pars_opts (int argc, char **argv)
 {
     int i;
@@ -1549,7 +1607,6 @@ int pars_opts (int argc, char **argv)
     char *max_str = 0;
     char *exclude_file = 0;
     char *allow_str = 0;
-    char **control_mirror_directories;
 
     struct prog_options arg_opts[] =
     {
@@ -1589,8 +1646,12 @@ int pars_opts (int argc, char **argv)
 	{0, "", "--no-mtimes", ARG_SET, 0, 0, &no_mtimes},
 	{0, "", "--no-chmod", ARG_CLEAR, 0, 0, &allow_chmod},
 	{0, "", "--no-chown", ARG_CLEAR, 0, 0, &allow_chown},
+	{0, "", "--no-netrc", ARG_CLEAR, 0, 0, &allow_netrc},
+	{0, "", "--netrc", ARG_SET, 0, 0, &allow_netrc},
 	{0, "", "--ignore-size", ARG_SET, 0, 0, &ignore_size},
+	{0, "", "--allow-empty-ftp-dirs", ARG_SET, 0, 0, &ftpfs_option_allow_empty_directories},
       	{'L', "", "--strict-locking", ARG_CLEAR, 0, 0, &only_read},
+        {'i', "", "--ignore-next-exclude", ARG_ADD, 0, 0, &ignore_next_exclude},
 	{0, 0, 0, 0, 0, 0, 0}
     };
     arg_opts[10].str = &exclude_file;
@@ -1643,8 +1704,8 @@ int pars_opts (int argc, char **argv)
     } else {
 	if (control_mirror_directories[2])
 	    return 1;
-	control_dir = pathdup (control_mirror_directories[0]);
-	mirror_dir = pathdup (control_mirror_directories[1]);
+	free_me[free_num++] = control_dir = pathdup (control_mirror_directories[0]);
+	free_me[free_num++] = mirror_dir = pathdup (control_mirror_directories[1]);
     }
 
 /* now apply some rules */
@@ -1711,15 +1772,38 @@ void message_callback (char *p, ...)
     }
 }
 
-char *anonymous_password_callback (void)
+char *anonymous_password_callback (char *host)
 {
-    return ftpfs_password;
-}
+    char *p;
 
-char *password_callback (char *a, char *b, char *c)
-{
+    /* password explicitly specified on commandline takes precedence */
     if (ftpfs_password)
 	return ftpfs_password;
+    /* then have a look at ~/.netrc */
+    if (allow_netrc
+	&& (p = lookup_netrc (host, "anonymous"))) {	/*EK */
+	if (verbose)
+	    verbmess ("found password for anonymous in ./netrc for", host);
+	return p;
+    }
+    return NULL;
+}
+
+char *password_callback (char *a, char *b, char *c, char *host, char *user)
+{				/*EK */
+    char *p;
+
+    /* password explicitly specified on commandline takes precedence */
+    if (ftpfs_password)
+	return ftpfs_password;
+    /* then have a look at ~/.netrc */
+    if (allow_netrc
+	&& (p = lookup_netrc (host, user))) {	/*EK */
+	if (verbose)
+	    verbmess ("found password in ./netrc for", host);
+	return p;
+    }
+    /* if everything else fails, then ask */
     return strdup (getpass (b));
 }
 
@@ -1769,6 +1853,9 @@ int main (int argc, char **argv)
 {
     char *p;
     int r;
+
+    memset (free_me, 0, sizeof (void *) * 256);
+
     p = strrchr (argv[0], '/');
     progname = strdup (p ? p + 1 : argv[0]);
 
@@ -1782,11 +1869,20 @@ int main (int argc, char **argv)
 	fprintf (stdout, "Mirrordir version " VERSION "\n");
 	exit (0);
     }
-    if (!strcmp (mirror_dir, control_dir)) {
-	progmess ("mirror and control directories are the same path", control_dir);
-	usage (1);
+    if (copy_mode) {
+	char **n;
+	for (n = control_files; *n; n++) {
+	    if (!strcmp (mirror_dir, *n)) {
+		progmess ("copying a path to the same path", *n);
+		usage (1);
+	    }
+	}
+    } else {
+	if (!strcmp (mirror_dir, control_dir)) {
+	    progmess ("mirror and control directories are the same path", control_dir);
+	    usage (1);
+	}
     }
-
     vfs_set_message_callback (message_callback);
     vfs_set_getpasswd_callback (password_callback);
     vfs_set_getanonpasswd_callback (anonymous_password_callback);
@@ -1810,6 +1906,15 @@ int main (int argc, char **argv)
 	sprintf (s, "%lukB", blocks_to_kilobytes (blocks_written));
 	infomess ("Total mirrored", s);
     }
+    free (progname);
+    if (control_mirror_directories)
+	free (control_mirror_directories);
+    while (free_num--)
+	if (free_me[free_num])
+	    free (free_me[free_num]);
+    free_all_hardlinks ();
+
+    mad_finalize (__FILE__, __LINE__);
 /* FIXME: should other return values be reported. Most people won't now how to read them anyway */
     if (r == -1 || real_bad_error) {
 	if (file_in_use == real_bad_error)
