@@ -54,7 +54,6 @@
 #   include <sys/select.h>
 #endif
 #include "fs.h"
-#include "mad.h"
 /* #include "../src/setup.h" */
 /* #include "../src/tty.h" */		/* enable/disable interrupt key */
 #include <netdb.h>		/* struct hostent */
@@ -87,6 +86,8 @@
 #    define MAXHOSTNAMELEN 64
 #endif
 
+#include "mad.h"
+
 #define UPLOAD_ZERO_LENGTH_FILE
 
 static int ftpfserrno;
@@ -150,10 +151,18 @@ static int __ftpfs_chdir (struct ftpfs_connection *bucket ,char *remote_path);
  * is supplied.
  * */
 
+#ifdef HAVE_MAD
+#define ftpfs_get_host_and_username(a,b,c,d,e) mad_ftpfs_get_host_and_username(a,b,c,d,e,__FILE__,__LINE__)
+char *mad_ftpfs_get_host_and_username (char *path, char **host, char **user, int *port, char **pass, char *file, int line)
+{
+    return mad_get_host_and_username (path, host, user, port, 21, 1, pass, file, line);
+}
+#else
 char *ftpfs_get_host_and_username (char *path, char **host, char **user, int *port, char **pass)
 {
     return get_host_and_username (path, host, user, port, 21, 1, pass);
 }
+#endif
 
 static int
 select_on_two (int fd1, int fd2)
@@ -415,7 +424,10 @@ login_server (struct ftpfs_connection *bucket, char *netrcpass)
     else {
         if (!strcmp (quser(bucket), "anonymous") || 
             !strcmp (quser(bucket), "ftp")) {
-	    op = strdup(ftpfs_anonymous_passwd);
+	    /* check if a specific anonymous password exists */
+	    op = load_anon_passwd (qhost(bucket)); /*EK*/
+	    /* if not, then use our default */
+	    if (!op) op = strdup(ftpfs_anonymous_passwd);
 	    anon = 1;
          } else {
             char *p;
@@ -423,7 +435,8 @@ login_server (struct ftpfs_connection *bucket, char *netrcpass)
 	    if (!bucket->password){
 		p = copy_strings (" FTP: Password required for ", quser(bucket), 
 				  " ", NULL);
-		op = input_dialog (p, "Password:", "");
+		op = input_dialog (p, "Password:", "", 
+				       qhost(bucket), quser(bucket)); /*EK*/
 		free (p);
 		if (op == NULL) {
 		    ftpfserrno = EPERM;
@@ -457,7 +470,8 @@ login_server (struct ftpfs_connection *bucket, char *netrcpass)
 	    wipe_password (proxypass);
 	p = copy_strings(" Proxy: Password required for ", proxyname, " ",
 			 NULL);
-	proxypass = input_dialog (p, "Password:", "");
+	proxypass = input_dialog (p, "Password:", "",
+					  NULL, NULL); /*EK*/
 	free(p);
 	if (proxypass == NULL) {
 	    ftpfserrno = EPERM;
@@ -623,8 +637,8 @@ load_no_proxy_list ()
 		continue;
 
 	    *p = '\0';
-	    p = xmalloc (strlen (s), "load_no_proxy_list:1");
-	    np = xmalloc (sizeof (*np), "load_no_proxy_list:2");
+	    p = malloc (strlen (s));
+	    np = malloc (sizeof (*np));
 	    strcpy (p, s);
 	    np->domain = p;
 	    np->next   = 0;
@@ -786,8 +800,7 @@ open_command_connection (char *host, char *user, int port, char *netrcpass)
     struct ftpfs_connection *bucket;
     int retry_seconds, count_down;
 
-    bucket = xmalloc(sizeof(struct ftpfs_connection), 
-		     "struct ftpfs_connection");
+    bucket = malloc(sizeof(struct ftpfs_connection));
     
     if (bucket == NULL) {
 	ftpfserrno = ENOMEM;
@@ -1147,18 +1160,25 @@ open_data_connection (struct ftpfs_connection *bucket, char *cmd, char *remote,
  * makes BUCKET point to the connection bucket descriptor for PATH
  * returns a malloced string with the pathname relative to BUCKET.
  */
-static char*
-ftpfs_get_path (struct ftpfs_connection **bucket, char *path)
+#ifdef HAVE_MAD
+#define ftpfs_get_path(a,b) mad_ftpfs_get_path(a,b,__FILE__,__LINE__)
+static char *mad_ftpfs_get_path (struct ftpfs_connection **bucket, char *path, char *file, int line)
+#else
+static char *ftpfs_get_path (struct ftpfs_connection **bucket, char *path)
+#endif
 {
     char *user, *host, *remote_path, *pass;
     int port;
 
     /* An absolute path name, try to determine connection socket */
-    if (strncmp (path, "ftp://", 6) == 0){
+    if (strncmp (path, "ftp://", 6) == 0) {
 	path += 6;
 
-	if (!(remote_path = ftpfs_get_host_and_username (path, &host, &user, 
-	    &port, &pass))) {
+#ifdef HAVE_MAD
+	if (!(remote_path = mad_ftpfs_get_host_and_username (path, &host, &user, &port, &pass, file, line))) {
+#else
+	if (!(remote_path = ftpfs_get_host_and_username (path, &host, &user, &port, &pass))) {
+#endif
 	    ftpfserrno = ENOENT;
 	    free (host);
 	    free (user);
@@ -1171,17 +1191,17 @@ ftpfs_get_path (struct ftpfs_connection **bucket, char *path)
 	    free (host);
 	    free (user);
 	    if (pass)
-	        wipe_password (pass);
+		wipe_password (pass);
 	    return NULL;
 	}
-        free (host);
-        free (user);
+	free (host);
+	free (user);
 	if (pass)
 	    wipe_password (pass);
 	return remote_path;
     }
     /* never get here !!! */
-    message_1s(1, " Error ", " Oops, you just hit a bug in the code ");
+    message_1s (1, " Error ", " Oops, you just hit a bug in the code ");
     return NULL;
 }
 
@@ -1262,8 +1282,7 @@ resolve_symlink(struct ftpfs_connection *bucket, struct ftpfs_dir *dir)
 		int r = strcmp(fe->name, filename);
 		free(filename);
 		if (r == 0) {
-		    fe->l_stat = xmalloc(sizeof(struct stat), 
-					 "resolve_symlink: struct stat");
+		    fe->l_stat = malloc(sizeof(struct stat));
 		    if (fe->l_stat == NULL)
 			goto done;
 		    *fe->l_stat = s;
@@ -1318,34 +1337,62 @@ ftpfs_same_dir (char *path, struct ftpfs_connection *bucket)
     return 0;
 }
 
+
+static int parse_buffer (struct linklist **file_list, char *buffer, struct ftpfs_connection *bucket, int has_symlinks)
+{
+    struct ftpentry *fe;
+    fe = malloc (sizeof (struct ftpentry));
+    bzero (fe, sizeof (struct ftpentry));
+    if (!fe)
+	return ENOMEM;
+    if (parse_ls_lga (buffer, &fe->s, &fe->name, &fe->linkname)) {
+	fe->count = 1;
+	fe->local_filename = fe->remote_filename = NULL;
+	fe->l_stat = NULL;
+	fe->bucket = bucket;
+	(fe->s).st_ino = bucket->__inode_counter++;
+	if (S_ISLNK (fe->s.st_mode))
+	    has_symlinks = 1;
+
+	if (!linklist_insert (*file_list, fe)) {
+	    free (fe);
+	    return ENOMEM;
+	}
+    } else
+	free (fe);
+    return 0;
+}
+
+int ftpfs_option_allow_empty_directories = 0;
+
 static struct ftpfs_dir *
-retrieve_dir(struct ftpfs_connection *bucket, char *remote_path)
+ retrieve_dir (struct ftpfs_connection *bucket, char *remote_path)
 {
 #ifdef OLD_READ
     FILE *fp;
 #endif
+    int err;
     int sock, has_symlinks;
     struct linklist *file_list, *p;
-    struct ftpentry *fe;
     char buffer[8192];
     struct ftpfs_dir *dcache;
     int got_intr = 0;
     int has_spaces = (strchr (remote_path, ' ') != NULL);
 
-    for (p = qdcache(bucket)->next;p != qdcache(bucket);
+    for (p = qdcache (bucket)->next; p != qdcache (bucket);
 	 p = p->next) {
 	dcache = p->data;
-	if (strcmp(dcache->remote_path, remote_path) == 0) {
+	if (strcmp (dcache->remote_path, remote_path) == 0) {
 	    struct timeval tim;
 
-	    gettimeofday(&tim, NULL);
+	    gettimeofday (&tim, NULL);
 	    if ((tim.tv_sec < dcache->timestamp.tv_sec) && !force_expiration)
 		return dcache;
 	    else {
 		force_expiration = 0;
 		p->next->prev = p->prev;
 		p->prev->next = p->next;
-		ftpfs_dir_destructor(dcache);
+		ftpfs_dir_destructor (dcache);
 		free (p);
 		break;
 	    }
@@ -1353,39 +1400,37 @@ retrieve_dir(struct ftpfs_connection *bucket, char *remote_path)
     }
 
     has_symlinks = 0;
-    print_vfs_message("ftpfs: Reading FTP directory...");
-    if (has_spaces) 
-        if (__ftpfs_chdir(bucket, remote_path) != COMPLETE) {
-            ftpfserrno = ENOENT;
-	    print_vfs_message("ftpfs: CWD failed.");
+    print_vfs_message ("ftpfs: Reading FTP directory...");
+    if (has_spaces)
+	if (__ftpfs_chdir (bucket, remote_path) != COMPLETE) {
+	    ftpfserrno = ENOENT;
+	    print_vfs_message ("ftpfs: CWD failed.");
 	    return NULL;
-        }
-
-    file_list = linklist_init();
+	}
+    file_list = linklist_init ();
     if (file_list == NULL) {
 	ftpfserrno = ENOMEM;
-	print_vfs_message("ftpfs: couldn't get a file listing");
-        return NULL;
-    }
-    dcache = xmalloc(sizeof(struct ftpfs_dir), 
-		     "struct ftpfs_dir");
-    if (dcache == NULL) {
-	ftpfserrno = ENOMEM;
-	linklist_destroy(file_list, NULL);
-	print_vfs_message("ftpfs: FAIL");
+	print_vfs_message ("ftpfs: couldn't get a file listing");
 	return NULL;
     }
-    gettimeofday(&dcache->timestamp, NULL);
+    dcache = malloc (sizeof (struct ftpfs_dir));
+    if (dcache == NULL) {
+	ftpfserrno = ENOMEM;
+	linklist_destroy (file_list, NULL);
+	print_vfs_message ("ftpfs: FAIL");
+	return NULL;
+    }
+    gettimeofday (&dcache->timestamp, NULL);
     dcache->timestamp.tv_sec += ftpfs_directory_timeout;
     dcache->file_list = file_list;
-    dcache->remote_path = strdup(remote_path);
+    dcache->remote_path = strdup (remote_path);
     dcache->count = 1;
 
     if (has_spaces)
-        sock = open_data_connection (bucket, "LIST -la", ".", 0);
+	sock = open_data_connection (bucket, "LIST -la", ".", 0);
     else {
 	char *path = copy_strings (remote_path, PATH_SEP_STR, ".", (char *) 0);
-	
+
 	sock = open_data_connection (bucket, "LIST -la", path, 0);
 	free (path);
     }
@@ -1395,103 +1440,100 @@ retrieve_dir(struct ftpfs_connection *bucket, char *remote_path)
 
 #ifdef OLD_READ
 #define close_this_sock(x,y) fclose (x)
-    fp = fdopen(sock, "r");
+    fp = fdopen (sock, "r");
     if (fp == NULL) {
 	ftpfserrno = errno;
-	close(sock);
-        goto error_2;
+	close (sock);
+	goto error_2;
     }
 #else
 #define close_this_sock(x,y) close (y)
 #endif
-    
+
     /* Clear the interrupt flag */
     enable_interrupt_key ();
-    
+
     errno = 0;
 #ifdef OLD_READ
     while (fgets (buffer, sizeof (buffer), fp) != NULL) {
 	if (got_intr = got_interrupt ())
 	    break;
 #else
-    while ((got_intr = get_line_interruptible (buffer, sizeof (buffer), sock)) != EINTR){
+    while ((got_intr = get_line_interruptible (buffer, sizeof (buffer), sock)) != EINTR) {
 #endif
 	int eof = got_intr == 0;
 
-	if (ftpfs_debug_server_dialog){
+	if (ftpfs_debug_server_dialog) {
 	    fputs (buffer, ftpfs_logfile);
 	    fflush (ftpfs_logfile);
 	}
-	if (buffer [0] == 0 && eof)
+	if (buffer[0] == 0 && eof)
 	    break;
-	fe = xmalloc(sizeof(struct ftpentry), "struct ftpentry");
-	bzero (fe, sizeof (struct ftpentry));
-	if (fe == NULL) {
-	    ftpfserrno = ENOMEM;
+	err = parse_buffer (&file_list, buffer, bucket, has_symlinks);
+	if (err == ENOMEM) {
+	    ftpfserrno = err;
 	    goto error_1;
 	}
-        if (parse_ls_lga (buffer, &fe->s, &fe->name, &fe->linkname)) {
-	    fe->count = 1;
-	    fe->local_filename = fe->remote_filename = NULL;
-	    fe->l_stat = NULL;
-	    fe->bucket = bucket;
-            (fe->s).st_ino = bucket->__inode_counter++;
-	    if (S_ISLNK(fe->s.st_mode))
-		has_symlinks = 1;
-	    
-	    if (!linklist_insert(file_list, fe)) {
-		free(fe);
-		ftpfserrno = ENOMEM;
-	        goto error_1;
-	    }
-	}
-	else
-	    free(fe);
 	if (eof)
 	    break;
     }
-    if (got_intr){
-	disable_interrupt_key();
-	print_vfs_message("ftpfs: reading FTP directory interrupt by user");
+    if (got_intr) {
+	disable_interrupt_key ();
+	print_vfs_message ("ftpfs: reading FTP directory interrupt by user");
 #ifdef OLD_READ
-	ftpfs_abort(bucket, fileno(fp));
+	ftpfs_abort (bucket, fileno (fp));
 #else
-	ftpfs_abort(bucket, sock);
+	ftpfs_abort (bucket, sock);
 #endif
-	close_this_sock(fp, sock);
+	close_this_sock (fp, sock);
 	ftpfserrno = EINTR;
 	goto error_3;
     }
-    close_this_sock(fp, sock);
-    disable_interrupt_key();
+    close_this_sock (fp, sock);
+    disable_interrupt_key ();
     if (getreply (qsock (bucket), NULL, 0) != COMPLETE) {
 	ftpfserrno = EIO;
-        goto error_3;
-    }
-    if (file_list->next == file_list) {
-	ftpfserrno = EACCES;
 	goto error_3;
     }
-    if (!linklist_insert(qdcache(bucket), dcache)) {
+    if (file_list->next == file_list) {
+	if (!ftpfs_option_allow_empty_directories) {
+	    ftpfserrno = EACCES;
+	    goto error_3;
+	} else {
+	    strcpy (buffer, "drwxrwxrwx   2 root     root         1024 Apr 17 12:38 .");
+	    err = parse_buffer (&file_list, buffer, bucket, has_symlinks);
+	    if (err == ENOMEM) {
+		ftpfserrno = err;
+		goto error_1;
+	    }
+	    strcpy (buffer, "drwxrwxrwx   2 root     root         1024 Apr 17 12:38 ..");
+	    err = parse_buffer (&file_list, buffer, bucket, has_symlinks);
+	    if (err == ENOMEM) {
+		ftpfserrno = err;
+		goto error_1;
+	    }
+	}
+    }
+    if (!linklist_insert (qdcache (bucket), dcache)) {
 	ftpfserrno = ENOMEM;
-        goto error_3;
+	goto error_3;
     }
     if (has_symlinks)
-	resolve_symlink(bucket, dcache);
-    print_vfs_message("ftpfs: got listing");
+	resolve_symlink (bucket, dcache);
+    print_vfs_message ("ftpfs: got listing");
     return dcache;
-error_1:
-    disable_interrupt_key();
-    close_this_sock(fp, sock);
+  error_1:
+    disable_interrupt_key ();
+    close_this_sock (fp, sock);
 #ifdef OLD_READ
-error_2: 
+  error_2:
 #endif
-    getreply(qsock(bucket), NULL, 0);
-error_3:
-    free(dcache->remote_path);
-    free(dcache);
-    linklist_destroy(file_list, ftpentry_destructor);
-    print_vfs_message("ftpfs: failed");
+    getreply (qsock (bucket), NULL, 0);
+  error_3:
+    free (dcache->remote_path);
+    free (dcache);
+    linklist_destroy (file_list, ftpentry_destructor);
+    print_vfs_message ("ftpfs: failed");
     return NULL;
 }
 
@@ -1615,7 +1657,7 @@ int retrieve_file_start2(struct ftpentry *fe)
 	return 0;
     }
     remotestat_size = fe->s.st_size;
-    remotebuffer = xmalloc (8192, "");
+    remotebuffer = malloc (8192);
     return 1;
 }
 
@@ -1721,9 +1763,9 @@ int ftpfs_setctl (char *path, int ctlop, char *arg)
         case MCCTL_SETTMPDIR:
 	    if (ftpfs_temp_name_dir)
 		free (ftpfs_temp_name_dir);
-	    if (arg)
+	    if (arg) {
 		ftpfs_temp_name_dir = strdup (arg);
-	    else
+	    } else
 		ftpfs_temp_name_dir = 0;
 	    return 1;
 
@@ -1741,7 +1783,11 @@ int retrieve_file(struct ftpentry *fe)
     if (fe->local_filename)
         return 1;
     fe->local_stat.st_mtime = 0;
+#ifdef HAVE_MAD
+    fe->local_filename = strdup(tempnam(ftpfs_temp_name_dir, "ftpfs"));
+#else
     fe->local_filename = tempnam(ftpfs_temp_name_dir, "ftpfs");
+#endif
     fe->local_is_temp = 1;
     if (fe->local_filename == NULL) {
 	ftpfserrno = ENOMEM;
@@ -1935,7 +1981,7 @@ _get_file_entry(struct ftpfs_connection *bucket, char *file_name,
     if ((op & FTPFS_OPEN) && (flags & O_CREAT)) {
 	int handle;
 
-	ent = xmalloc(sizeof(struct ftpentry), "struct ftpentry");
+	ent = malloc(sizeof(struct ftpentry));
 	bzero (ent, sizeof (struct ftpentry));
 	if (ent == NULL) {
 	    ftpfserrno = ENOMEM;
@@ -2002,6 +2048,7 @@ static int remove_temp_file (char *file_name)
     q = *p;
     *p = '\0';
     dcache = retrieve_dir (bucket, *file_name ? file_name : "/");
+    free (file_name);
     if (dcache == NULL)
 	return -1;
     file_list = dcache->file_list;
@@ -2040,9 +2087,11 @@ static int flush_dir (char *remote_path)
 	    p->prev->next = p->next;
 	    ftpfs_dir_destructor (dcache);
 	    free (p);
+	    free (remote_path);
 	    return 0;
 	}
     }
+    free (remote_path);
     return -1;
 }
 
@@ -2103,6 +2152,8 @@ send_ftp_command(char *filename, char *cmd, int flags)
     return 0;
 }
 
+static char *free_passwd = 0;
+
 /* This routine is called as the last step in load_setup */
 void
 ftpfs_init_passwd(void)
@@ -2111,10 +2162,12 @@ ftpfs_init_passwd(void)
     char *p, hostname[MAXHOSTNAMELEN];
     struct hostent *hp;
 
-    ftpfs_anonymous_passwd = load_anon_passwd ();
+    /* establish if there is an anonymous password */
+    ftpfs_anonymous_passwd = load_anon_passwd ("*"); /*EK*/
     if (ftpfs_anonymous_passwd)
 	return;
 
+    /* if there isn't, then make one */
     if ((passwd_info = getpwuid (geteuid ())) == NULL)
 	p = "guest";
     else
@@ -2125,6 +2178,7 @@ ftpfs_init_passwd(void)
 	ftpfs_anonymous_passwd = copy_strings (p, "@", hp->h_name, NULL);
     else
 	ftpfs_anonymous_passwd = copy_strings (p, "@", hostname, NULL);
+    free_passwd = ftpfs_anonymous_passwd;
     endpwent ();
 }
 
@@ -2141,6 +2195,10 @@ ftpfs_done(void)
     linklist_destroy(ftpfs_connections_list, ftpfs_connection_destructor);
     if (ftpfs_debug_server_dialog)
 	fclose (ftpfs_logfile);
+    if (free_passwd)
+	wipe_password (free_passwd);
+    if (ftpfs_temp_name_dir)
+	free (ftpfs_temp_name_dir);
 }
 
 /* The callbacks */
@@ -2156,7 +2214,7 @@ static void *ftpfs_open (char *file, int flags, int mode)
     struct ftpfs_filp *fp;
     struct ftpentry *fe;
 
-    fp = xmalloc(sizeof(struct ftpfs_filp), "struct ftpfs_filp");
+    fp = malloc(sizeof(struct ftpfs_filp));
     if (fp == NULL) {
 	ftpfserrno = ENOMEM;
         return NULL;
@@ -2280,7 +2338,7 @@ static void *ftpfs_opendir (char *dirname)
 
     if (!(remote_path = ftpfs_get_path (&bucket, dirname)))
         return NULL;
-    dirp = xmalloc(sizeof(struct ftpfs_dirent), "struct ftpfs_dirent");
+    dirp = malloc(sizeof(struct ftpfs_dirent));
     if (dirp == NULL) {
 	ftpfserrno = ENOMEM;
 	goto error_return;
@@ -2730,7 +2788,7 @@ int lookup_netrc (char *host, char **login, char **pass)
                 *pass = strdup (rupp->pass);
             return 0;
         }
-    netrcname = xmalloc (strlen (home_dir) + strlen ("/.netrc") + 1, "netrc");
+    netrcname = malloc (strlen (home_dir) + strlen ("/.netrc") + 1);
     strcpy (netrcname, home_dir);
     strcat (netrcname, "/.netrc");
     netrcp = netrc = load_file (netrcname);
@@ -2814,7 +2872,7 @@ int lookup_netrc (char *host, char **login, char **pass)
 	else
 	    break;
     }
-    rupp = (struct rupcache *) xmalloc (sizeof (struct rupcache), "");
+    rupp = (struct rupcache *) malloc (sizeof (struct rupcache));
     rupp->host = strdup (host);
     rupp->login = rupp->pass = 0;
     
