@@ -186,7 +186,8 @@ static int mcfs_login_server (int my_socket, char *user, int port,
 	if (netrcpass != NULL)
 	    pass = strdup (netrcpass);
 	else
-	    pass = input_dialog (" MCFS Password required ", "Password:", "");
+	    pass = input_dialog (" MCFS Password required ", "Password:", "",
+				  NULL, NULL); /*EK*/
 	if (!pass){
 	    rpc_send (my_socket, RPC_INT, MC_QUIT, RPC_END);
 	    close (my_socket);
@@ -498,7 +499,7 @@ static void *mcfs_open (char *file, int flags, int mode)
     if (is_error (result, error_num))
 	return 0;
 
-    remote_handle = (mcfs_handle *) xmalloc (2 * sizeof (mcfs_handle), "mcfs_handle");
+    remote_handle = (mcfs_handle *) malloc (2 * sizeof (mcfs_handle));
     remote_handle->handle  = result;
     remote_handle->conn    = mc;
     
@@ -613,7 +614,7 @@ static void *mcfs_opendir (char *dirname)
 	    
     handle = result;
 
-    mcfs_info = (opendir_info *) xmalloc (sizeof(opendir_info),"mcfs_opendir");
+    mcfs_info = (opendir_info *) malloc (sizeof(opendir_info));
     mcfs_info->conn = mc;
     mcfs_info->handle = handle;
     mcfs_info->entries = 0;
@@ -643,8 +644,8 @@ static int mcfs_loaddir (opendir_info *mcfs_info)
 	if (entry_len == 0)
 	    break;
 
-	new_entry = xmalloc (sizeof (dir_entry), "mcfs_loaddir");
-	new_entry->text = xmalloc (entry_len + 1, "mcfs_loaddir");
+	new_entry = malloc (sizeof (dir_entry));
+	new_entry->text = malloc (entry_len + 1);
 	new_entry->text [entry_len] = 0;
 
 	new_entry->next = 0;
