@@ -215,13 +215,13 @@ void vfs_addstamp (vfs *v, vfsid id, struct vfs_stamping *parent)
 		gettimeofday(&(stamp->time), NULL);
                 return;
 	    }
-        stamp = xmalloc (sizeof (struct vfs_stamping), "vfs stamping");
+        stamp = malloc (sizeof (struct vfs_stamping));
         stamp->v = v;
         stamp->id = id;
 	if (parent) {
 	    struct vfs_stamping *st = stamp;
 	    for ( ; parent; ) {
-		st->parent = xmalloc (sizeof (struct vfs_stamping), "vfs stamping");
+		st->parent = malloc (sizeof (struct vfs_stamping));
 		*st->parent = *parent;
 		parent = parent->parent;
 		st = st->parent;
@@ -392,7 +392,7 @@ DIR *mc_opendir (char *dirname)
 
     if (dirname [i - 1] != '/') { 
     /* We should make possible reading of the root directory in a tar file */
-        p = xmalloc (i + 2, "slash");
+        p = malloc (i + 2);
         strcpy (p, dirname);
         strcpy (p + i, "/");
         dirname = p;
@@ -416,7 +416,7 @@ DIR *mc_opendir (char *dirname)
     if (p)
         free (p);
     
-    handlep = (int *) xmalloc (sizeof (int), "opendir handle");
+    handlep = (int *) malloc (sizeof (int));
     *handlep = handle;
     return (DIR *) handlep;
 }
@@ -1137,7 +1137,7 @@ int mc_chdir (char *path)
 
     if (path [i - 1] != '/') {
     /* We should make possible reading of the root directory in a tar archive */
-        p = xmalloc (i + 2, "slash");
+        p = malloc (i + 2);
         strcpy (p, path);
         strcpy (p + i, "/");
         path = p;
@@ -1240,7 +1240,7 @@ char *vfs_get_current_dir (void)
 
 static void vfs_setup_wd (void)
 {
-    current_dir = xmalloc (MC_MAXPATHLEN, "Current directory");
+    current_dir = malloc (MC_MAXPATHLEN);
     if (!get_current_wd (current_dir, MC_MAXPATHLEN))
 	if (errno != 0)
 	     current_dir [0] = 0;
@@ -1320,7 +1320,7 @@ caddr_t mc_mmap (caddr_t addr, size_t len, int prot, int flags, int fd, off_t of
 	errno = (*vfs->ferrno)();
 	return (caddr_t)-1;
     }
-    mcm = (struct mc_mmapping *) xmalloc (sizeof (struct mc_mmapping), "vfs: mmap handling");
+    mcm = (struct mc_mmapping *) malloc (sizeof (struct mc_mmapping));
     mcm->addr = result;
     mcm->vfs_info = vfs_info (fd);
     mcm->vfs = vfs;
@@ -1871,7 +1871,7 @@ int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname)
 	    
 	if (filename) {
 	    p = column_ptr [idx2] - column_ptr [idx];
-	    s = xmalloc (p, "filename");
+	    s = malloc (p);
 	    strncpy (s, p_copy + column_ptr [idx], p - 1);
 	    s[p - 1] = '\0';
 	    *filename = s;
